#include "vieter_tree_balancing.h"
#include <string.h>

bool vieter_tree_node_validate(vieter_tree_node *node,
                               uint64_t passed_black_nodes,
                               uint64_t expected_black_nodes) {
  // Path to NULL children should contain the same amount of black nodes
  if (node == NULL) {
    return passed_black_nodes == expected_black_nodes;
  }

  if (vieter_tree_node_get(node, vieter_tree_node_black)) {
    passed_black_nodes++;
  }

  // Either the node itself is black, or its children are both either NULL or
  // black.
  bool correctly_colored_children =
      vieter_tree_node_get(node, vieter_tree_node_black) ||
      ((node->children[0] == NULL ||
        vieter_tree_node_get(node->children[0], vieter_tree_node_black)) &&
       (node->children[1] == NULL ||
        vieter_tree_node_get(node->children[1], vieter_tree_node_black)));

  return correctly_colored_children &&
         vieter_tree_node_validate(node->children[0], passed_black_nodes,
                                   expected_black_nodes) &&
         vieter_tree_node_validate(node->children[1], passed_black_nodes,
                                   expected_black_nodes);
}

vieter_tree_node *vieter_tree_node_balance(vieter_tree_node *node) {
  vieter_tree_node *parent = node->parent;
  vieter_tree_node *grand_parent = parent->parent;
  vieter_tree_node *root = grand_parent;
  vieter_tree_node *children[2];
  vieter_tree_node *grand_children[4];

  uint64_t key_root = root->key;
  void *data_root = root->data;

  if (vieter_tree_node_get(node, vieter_tree_node_right)) {
    children[0] = parent;
    children[1] = node;

    if (vieter_tree_node_get(parent, vieter_tree_node_right)) {
      root->key = parent->key;
      root->data = parent->data;

      parent->key = key_root;
      parent->data = data_root;

      grand_children[0] = grand_parent->children[0];
      grand_children[1] = parent->children[0];
      grand_children[2] = node->children[0];
      grand_children[3] = node->children[1];
    } else {
      root->key = node->key;
      root->data = node->data;

      node->key = key_root;
      node->data = data_root;

      grand_children[0] = parent->children[0];
      grand_children[1] = node->children[0];
      grand_children[2] = node->children[1];
      grand_children[3] = grand_parent->children[1];
    }
  } else {
    children[0] = node;
    children[1] = parent;

    if (vieter_tree_node_get(parent, vieter_tree_node_right)) {
      root->key = node->key;
      root->data = node->data;

      node->key = key_root;
      node->data = data_root;

      grand_children[0] = grand_parent->children[0];
      grand_children[1] = node->children[0];
      grand_children[2] = node->children[1];
      grand_children[3] = parent->children[1];
    } else {
      root->key = parent->key;
      root->data = parent->data;

      parent->key = key_root;
      parent->data = data_root;

      grand_children[0] = node->children[0];
      grand_children[1] = node->children[1];
      grand_children[2] = parent->children[1];
      grand_children[3] = grand_parent->children[1];
    }
  }

  vieter_tree_node_set(
      root, vieter_tree_node_right,
      vieter_tree_node_get(grand_parent, vieter_tree_node_right));

  vieter_tree_node_set(children[0], vieter_tree_node_black, true);
  vieter_tree_node_set(children[1], vieter_tree_node_black, true);

  vieter_tree_node_set(root, vieter_tree_node_black, false);

  vieter_tree_node_set_children(children[0], grand_children);
  vieter_tree_node_set_children(children[1], grand_children + 2);
  vieter_tree_node_set_children(root, children);

  return root;
}

void vieter_tree_node_balance_after_insert(vieter_tree_node *node) {
  while (!vieter_tree_node_get(node, vieter_tree_node_black) &&
         node->parent != NULL &&
         !vieter_tree_node_get(node->parent, vieter_tree_node_black)) {
    node = vieter_tree_node_balance(node);
  }

  // The root must always be black
  if (node->parent == NULL) {
    vieter_tree_node_set(node, vieter_tree_node_black, true);
  }
}

void vieter_tree_node_remove_balanced(vieter_tree_node *node) {
  // A red node can only have 0 or 2 children. The node we receive only has
  // one child at most, so we know if it's red that it doesn't have any
  // children. A black node that has a single (right) child can be replaced by
  // this child, granted it becomes black as well. Either way, the node can be
  // replaced by its right child (even if it's NULL).
  if (!vieter_tree_node_get(node, vieter_tree_node_black) ||
      node->children[1] != NULL) {
    vieter_tree_node_set_child(
        node->parent, node->children[1],
        vieter_tree_node_get(node, vieter_tree_node_right));

    if (node->children[1] != NULL)
      vieter_tree_node_set(node->children[1], vieter_tree_node_black, true);

    vieter_tree_node_free(node);

    return;
  }

  // The complicated case is when we want to remove a black leaf
}
